/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                         Refresh.c                       **/
/**                                                         **/
/** This file contains screen refresh drivers which are     **/
/** used to convert the contents of VRAM[] and VDP[] to     **/
/** an ximage in zpixmap format using 1,2,4,8,16,24 or 32   **/
/** bits per pixels.                                        **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995,1996            **/
/**               Arnold Metselaar 1998                     **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify us, if you make any    **/
/**     changes to this file.                               **/
/**---------------------------------------------------------**/
/** Adaptions for any X-terminal by Arnold Metselaar.       **/
/**     If you want to use this file for VGA, you will need **/
/**     to adapt MSDOS.c . Please, notify me as well, if    **/
/**     you make any changes to this file.                  **/
/*************************************************************/

/** get Type for storing X-colors **/
#include <string.h>
#include <X11/Xlib.h>
#include "MSX.h"
#include "AnyX.h"
#define INLINE 

/** convenience macros to read VDP-registers **/
#define SP2 (VDP[25]&0x01)
#define YJK (VDP[25]&0x08)
#define YAE (VDP[25]&0x10)
#define EVEN_ODD (VDP[9]&4)
#define INTLAC   (VDP[9]&8)
#define INTERLACING (INTLAC&&EVEN_ODD&&(VDP[2]&0x020))

/* Generate several instances of some functions to cut down the number
    of function pointer dereferences */
#define BLOATIFY(name, proto)			\
name##N(byte)					\
name##N(word)					\
name##N(long)					\
name##W(byte)					\
name##W(word)					\
name##W(long)					\
static void (*name[6])proto = {			\
  name##N_byte, name##W_byte,   		\
  name##N_word, name##W_word,	 		\
  name##N_long, name##W_long, 			\
} ;

#define XBLOATIFY(name, proto)			\
name##X(byte,N,*buf,buf++)			\
name##X(word,N,*buf,buf++)			\
name##X(long,N,*buf,buf++)			\
name##X(byte,W,buf[0]=buf[1],buf+=2)		\
name##X(word,W,buf[0]=buf[1],buf+=2)		\
name##X(long,W,buf[0]=buf[1],buf+=2)		\
static void (*name[6])proto = {			\
  name##N_byte, name##W_byte,   		\
  name##N_word, name##W_word,			\
  name##N_long, name##W_long, 			\
} ;

/* Buffer to hold a line from the image before it is converted to a
    format with bitpix \not\in {8,16,32} */
static ColTyp linbuf[512];

typedef void (*pptyp)(byte*,int);

typedef struct  
{ 
  int bp_bitpix; 
  int bp_index;
  /* void (* bp_postproc)(byte *, int); */
  pptyp bp_postproc;
} bp_info; 

/** description of a sprite attribute entry **/
typedef struct
{
    byte spr_y, spr_x, spr_nr, spr_col ;
} spr_struct;

/*** Some variables that are related to the MSX-screen ***/
int NSprMax_M=4, NSprMax_C=8; 
static struct SprLin {
    byte plane,line; } SprBuf[32];
static byte SBuf_[257];
#define SBuf (SBuf_+1)
static const byte SprCol8[16] = { 0,1,12,13,96,97,108,109, 
				  157,3,28,31,224,227,252,255 };

/* Some macros representing the way the VDP maps table-elements to
 * addresses in the VRAM. This mapping may look weird but it saves an
 * addition in the real VDP. 
 *
 * Schematicly the mapping looks like this:
 * ...rrrr..r111...   (address-mask from VDP-registers)
 * ...11nn..nnnn...   ('logical' address within the table)
 * ---------------- &
 * ...rr**..*nnn...   (Address in the VRAM)
 * 
 * Some tables 'fit in their masks' and their mapping is equivalent to
 * an addition:
 * ...rrrr1111...     ...rrrr0000...   (address from VDP-registers)
 * ...1111nnnn...     ...0000nnnn...   ('logical' address within the table)
 * --------------&    --------------+
 * ...rrrrnnnn...     ...rrrrnnnn...   (Address in the VRAM)
 *
 * The following bits in the VDP-registers are related to alternating
 * and interlacing:
 *   VDPStatus[2]&2: EO_     - Display field flag (0: display first field)
 *      (not implemented in this emulator)
 *   VDP[9]&4:       EO      - Even or Odd Display  (0: one page; 1: two pages)
 *   VDP[9]&8:       IL      - Interlace
 * When EO is set, there is an extra mask in the mapping for half of
 * the screen refreshes.  When IL is set as well, the vertical
 * resolution is effectively doubled.
 *
 * I assume that there is only an extra mask for the name table and
 * that it is 0x0ffff in screen 7 and 8 and 0x17fff in all other
 * modes. This comes down to the same bit in VDP[2]
 * 
 * In the comments before functions using tables:
 *   stp:  smallest possible change in adderss from vdp-registers
 *   len:  smallest positive logical address that does not fit in the n-bits
 *   xmsk: extra mask used for alternate display (EO)
 *           if xmsk is not specified, it is not used
 *   siz:  number of bytes in table that can be used (for scrolled area)
 *           a range may be specified instead of siz
 */

/* Convenience macros for the mapping of table elements to VRAM positions */
#define ADDR_MSK(r,sh) ((((long)VDP[r]+1)<<(sh))-1)
#define ADDR_MSK2(r1,r2,sh) (((((long)VDP[r2]<<8)+VDP[r1]+1)<<(sh))-1)
#define T_OFF(n,msk,len) ((msk)&(((1L<<17)-(len))|(n)))

/* name table, mapped differently for screen 7 and 8 , influenced by
   even or odd display and by two page horizontal scroll: 
   p!=0 means display the even page (force bit 5 of VDP[2] to zero)
   This is used for both even/odd display (interlacing) and 2-page
   horizontal scroll 
*/
#define CHRTAB_MSK(p,sh) ((((long)(VDP[2]&((p)?0xdf:0xff))+1)<<(sh))-1)
#define CHRTAB(n,p,len) T_OFF((n),CHRTAB_MSK(p,10),(len))
#define CHRTAB78(n,p,len) T_OFF((n),CHRTAB_MSK(p,11),(len))

#define NEXTCHAR(lx,T,base,p,len) \
T=( (++lx & 0x1f) ? T : \
    VRAM+CHRTAB(base, p || ( ( (lx&0x20)==0) && (SP2) ), len))

#define NEXTCHAR78(lx,T,base,p,len) \
T=( (++lx & 0x1f) ? T : \
    VRAM+CHRTAB78(base, p || ( ((lx&0x20)==0) && (SP2) ), len))

/* #define CHRTAB78(n,len) T_OFF(n,CHRTAB78_MSK,len) */
/* color table */
#define COLTAB_MSK ADDR_MSK2(3,10,6)
#define COLTAB(n,len) T_OFF(n,COLTAB_MSK,len)
/* character pattern table */
#define CHRGEN_MSK ADDR_MSK(4,11)
#define CHRGEN(n,len) T_OFF(n,CHRGEN_MSK,len)
/* sprite attribute table */
#define SPRTAB_MSK ADDR_MSK2(5,11,7)
#define SPRTAB(n,len) T_OFF(n,SPRTAB_MSK,len)
/* sprite patterns, just for completeness. these definitions are not
   used because this table fits in its mask */
#define SPRGEN_MSK ADDR_MSK(6,11)
#define SPRGEN(n,len) T_OFF(n,SPRGEN_MSK,len) 

int eo_flag; /* set if it is time to display the alternate field */

/* Macro to pass needed Palette-entries to functions in Colors.c: */
#define INIPAL(colset) 				\
{						\
  int ColSet;					\
  ColSet=(ScreenON?(colset):1<<BGColor);	\
  if (PalOut & ColSet) Colors.Take(ColSet);	\
  PalUsed|=ColSet;				\
}

/* index of the PixLine<mode>() routine that should be used: */
static int PL_ind; 
/* pointer to a routine that will copy one line and convert it on the
   fly: */
static pptyp postproc;
/* static void (* postproc)(byte *, int); */


static byte *
ScrPtr(int y)
{
  return XBuf+(long)WIDTH*(high?2:1)*y*bitpix/8;
}

static byte *
ScrPtr_Odd(int y)
{
  return XBuf+(long)WIDTH*(2*y+1)*bitpix/8;
}

/** Some routines to copy a line from linbuf to Xbuf, correcting the
    number of bits per pixels on the fly **/
static void 
cp32_24(byte * dest, int len)
{
  byte * D;
  ColTyp *S;

  len=wide?2*len:len;
  D=dest;
  for( S=linbuf ; S<linbuf+len ; S++)
    {
      *(D++)=(*S).ct_Byte[0];
      *(D++)=(*S).ct_Byte[1];
      *(D++)=(*S).ct_Byte[2];
    }
}
static void 
cp8_4m(byte * dest, int len)
{
  byte *D, *S, *last;

  len=wide?2*len:len;
  D=dest;
  for( S=(char*)linbuf, last=S+len ; S<last ; S+=2)
    {
      *(D++)=S[0]*16 + S[1];
    }
}
static void 
cp8_2m(byte * dest, int len)
{
  byte *D, *S, *last;

  len=wide?2*len:len;
  D=dest;
  for( S=(char*)linbuf, last=S+len ; S<last ; S+=4)
    {
      *(D++)=S[0]*64 + S[1]*16 + S[2]*4 + S[3];
    }
}
static void 
cp8_1m(byte * dest, int len)
{
  byte *D, *S, *last;

  len=wide?2*len:len;
  D=dest;
  for( S=(char*)linbuf, last=S+len ; S<last ; S+=8)
    {
      *(D++)=S[0]*128 + S[1]*64 + S[2]*32 + S[3]*16
	+ S[4]*8 + S[5]*4 + S[6]*2 + S[7];
    }
}
static void 
cp8_4l(byte * dest, int len)
{
  byte *D, *S, *last;

  len=wide?2*len:len;
  D=dest;
  for( S=(char*)linbuf, last=S+len ; S<last ; S+=2)
    {
      *(D++)=S[0] + S[1]*16;
    }
}
static void 
cp8_2l(byte * dest, int len)
{
  byte *D, *S, *last;

  len=wide?2*len:len;
  D=dest;
  for( S=(char*)linbuf, last=S+len ; S<last ; S+=4)
    {
      *(D++)=S[0] + S[1]*4 + S[2]*16 + S[3]*64;
    }
}
static void 
cp8_1l(byte * dest, int len)
{
  byte *D, *S, *last;

  len=wide?2*len:len;
  D=dest;
  for( S=(char*)linbuf, last=S+len ; S<last ; S+=8)
    {
      *(D++)=S[0] + S[1]*2 + S[2]*4 + S[3]*8
	+ S[4]*16 + S[5]*32 + S[6]*64 + S[7]*128;
    }
}

/** Pixel setting routines for any number of bits per pixel **/
bp_info PixSetters[10]= 
{
/* for msb first: */
  { 1, 0, cp8_1m },
  { 2, 0, cp8_2m },
  { 4, 0, cp8_4m },
/* for lsb first: */ 
  { 0x81, 0, cp8_1l },
  { 0x82, 0, cp8_2l },
  { 0x84, 0, cp8_4l },
/* Whole bytes, same routines for MSB-first and LSB-first */
  { 8, 0, (void (*)(byte*,int)) 0 },
  {16, 2, (void (*)(byte*,int)) 0 },
  {24, 4, cp32_24 },
  {32, 4, (void (*)(byte*,int)) 0 }
};

/* Some convenient functions that probably should be inlined */
/* called from RefreshLine[AC]: */
INLINE static void
getJK(byte *p, int *J, int *K)
{
  *K=(p[0]&0x07) | ((p[1]&0x07)<<3);
  if (*K&0x20)  *K-=64;
  *J=(p[2]&0x07) | ((p[3]&0x07)<<3);
  if (*J&0x20)  *J-=64;
}

INLINE static ColTyp 
YJKcolor(int Y, int J, int K)
{
  int r,g,b;
  byte pal2p[3];
		
  r=Y+J;
  g=Y+K;
  b=((5*Y-2*J-K)/4);
  pal2p[0]=(r<0) ? 0 : ((r>31)?31:r);
  pal2p[1]=(g<0) ? 0 : ((g>31)?31:g);
  pal2p[2]=(b<0) ? 0 : ((b>31)?31:b);
  return Colors.GetCol2p(pal2p);
}

/* duplicates scanlines, used if interlacing is off  */
static void 
NoIntLac(register byte Y)
{
  register byte *P;
  register word linlen;
  
  linlen=WIDTH*bitpix/8;
  P=XBuf+(long)2*Y*linlen;
  memcpy(P+linlen,P,linlen);
}

/** Function to scan the sprite-info-table. NSpr and SprBuf
 * will be set as follows:
 * return value      = #sprites on screen line V
 * SprBuf[x-1].plane = planenumber of the x-th sprite on line V
 * SprBuf[x-1].line  = line within the x-th sprite that falls 
 *                           on screen line V
 * Table usage:
 *   sprite attribute table (sprtab, stp=0x80, len=0x80) :  0x0..0x7F 
 * OR
 *   sprite attribute table (sprtab, stp=0x80, len=0x400) : 0x200..0x27F
 *
 * The value of SprTab set in MSX.c happens to be correct in both cases
 */
static int 
ScanSprites (register int V, register int brkval, register int NSprMax)
{
  register byte H,N,sc,ys;
  register spr_struct *AT;
  int cnt,maxcnt;

  sc=BigSprites?2:1;
  H=(Sprites16x16?16:8);
  maxcnt=NSprMax;
  AT=(spr_struct *)SprTab;
  for ( N=cnt=0 ; (N < 32) && (cnt < maxcnt) ; N++)
    {
      if (AT[N].spr_y==brkval) break;
      if (AT[N].spr_y==brkval+1) continue;
      
      ys=V-(AT[N].spr_y+1);
      ys/=sc;
      if (ys<H)
	{
	  SprBuf[cnt].plane=N;
	  SprBuf[cnt++].line=ys;
	}
    }
  return cnt;
}

/* Function that sets SBuf according to the sprites on line V,
 * used for screen 4,5,6,7,8,10/11,12 
 * Returns non-zero if there is at least one sprite on line V.
 * Table usage: 
 *   sprite colour  table (sprtab, stp= 0x80, len=0x400): 0..0x1FF
 *   sprite attrib. table (sprtab, stp= 0x80, len=0x400): 0x200..0x27F
 *   sprite pattern table (sprgen, stp=len=siz=0x800)
 */
static int 
ColorSprites(register byte V)
{
  register byte C,H,M,N,sc;
  register spr_struct *AT;
  register word K;
  int L,R,NSpr;
  byte LinBuf[258];
  
  NSpr=ScanSprites(V,216,NSprMax_C);
  if (NSpr==0) return 0;
  memset(SBuf,0,256);

  H=Sprites16x16?0xFC:0xFF;
  M=SolidColor0?0x10:0;
  sc=BigSprites?2:1;
  AT=(spr_struct *)SprTab;
  memset(LinBuf+1,0,256);
  memset(SBuf,0,256);
  for ( N=NSpr,R=0; N--; )
    {
      C=VRAM[SPRTAB(((word)SprBuf[N].plane<<4)+SprBuf[N].line, 0x400)];
      if (M|(C&0x0F))
	{
	  spr_struct spr;
	  byte *P;
	  C|=M;
	  spr=AT[SprBuf[N].plane];
	  L=spr.spr_x+((C&0x80)?-31:1)
	    -VDP[27]; /* compensate for VDP[27] scrolling, 
			 sprites don't scroll */
	  P=SprGen+(((int)(spr.spr_nr & H)<<3)+SprBuf[N].line);
	  K=(((word)*P)<<8) + ((H==0xFC)?P[16]:0);
	  if (L<0)
	    {
	      int t;
	      t=(1-L)/sc;
	      K<<=t; L+=sc*t;
	    }
	  for (  ; K&&(L<257) ; K<<=1, L+=sc)
	    if (K&0x8000)
	      {
		LinBuf[L]|=C;
		if (sc==2) LinBuf[L+1]|=C;
	      }
	}
      if (!(C&0x40))
	{
	  R++;
	  for ( L=VDP[27] ; L<256 ; L++)
	    if (LinBuf[1+L]) 
	      SBuf[L]=LinBuf[1+L]; 
	  memset(LinBuf+1,0,256);
	}
    }
  return R;
}

/* Function that sets SBuf according to the sprites on line V,
 * used for screen 1,2,3. This function may give incorrect results
 * if sprites are overlapping and have bit 6 set in the color-byte. 
 * Table usage: 
 *   sprite attrib. table (sprtab, stp=len=siz= 0x80)
 *   sprite pattern table (sprgen, stp=len=siz=0x800)
 */
static int 
MonoSprites(register byte V)
{
  register byte C,H,M,sc,*P;
  register spr_struct *AT,*SP;
  register word K;
  short int L,N,NSpr;
  int SMin, R;

  NSpr=ScanSprites(V,208,NSprMax_M);
  if (NSpr==0) return 0;
  memset(SBuf,0,256);

  AT=(spr_struct *)SprTab;
  for (SMin=0 ; (SMin<NSpr)&&(AT[SprBuf[SMin].plane].spr_col&0x40) ; SMin++)
      ;
  H=Sprites16x16?0xFC:0xFF;
  M=SolidColor0?0x10:0;
  sc=BigSprites?2:1;
  for (N=NSpr-1,R=0 ; N>=SMin ; N--)
  {
    SP=AT+(SprBuf[N].plane);
    if (M|(C=SP->spr_col))
    {
      L=SP->spr_x-((C&0x80)?32:0)-VDP[27];/* compensate for VDP[27] scrolling;
					     sprites don't scroll */
      C|=M;
      P=SprGen+((int)(SP->spr_nr&H)<<3)+SprBuf[N].line;
      K=(((word)*P)<<8) + ((H==0xFC)?P[16]:0);
      if (L<0)
	{
	  int t;
	  t=(-L)/sc;
	  K<<=t; L+=sc*t;
	}
      if (K&&(L<256)) 
	R++;
      /* L can be -1 but SBuf is #defined as (S_buf+1) so that is no
         problem */
      for ( ; K&&(L<256); K<<=1, L+=sc)
        if (K&0x8000)
        {
          SBuf[L]=C;
          if ((sc==2) && (L+1<256)) SBuf[L+1]=C;
        };
    } 
  }
  return R;
}

/*  SCREEN 0: TXT 40x24/26.5(*)
 *  VDP[23] scrolls blocks of 8 screen-lines (one textline), VDP[26] is ignored
 *  Table usage:
 *    name    table (chrtab, stp=len=0x400, xmsk=0x17FFF (?), siz=0x3C0/0x438)
 *    pattern table (chrgen, stp=len=siz=0x800, xmsk=0x0FFFF (?) )
 *  Sprites: none
 *
 * (*): Having 26.5 text-lines in Text 40 is not very useful because
 *      the last 0x38 elements of the name table are always taken from the
 *      same VRAM locations as the first 0x38 elements.  
 */
#define PixLine0X(type,X,dest,inc)     				\
static void 							\
PixLine0##X##_##type(int p, byte Y, byte *S, void *Buf)		\
{								\
  long MSK, T;							\
  int X,J,K;							\
  type fc,bc,*buf;						\
								\
  buf=(type*)Buf;						\
  bc=XPal[BGColor].ct_##type;					\
  fc=(FGColor||SolidColor0)?(XPal[FGColor].ct_##type):bc;	\
  MSK= ~CHRTAB_MSK(p,10);					\
  T=CHRTAB((Y>>3)*40, p, 0x400);				\
  for ( X=40;X;X--)						\
    {								\
      if (T&MSK) T=CHRTAB((Y>>3)*40+(40-X), p, 0x400);		\
      K=*(S+((long)VRAM[T++]<<3));				\
      for ( J=0x80 ; J>2 ; J>>=1,inc)  dest=((K&J)?fc:bc);	\
    }								\
}
XBLOATIFY(PixLine0,(int p, byte Y, byte *S, void *Buf))

void RefreshLine0(register byte Y)
{
  ColTyp BC;
  int p;
  register byte *S;
  
  if (!Mapped) return;
  INIPAL((1<<BGColor)|(1<<FGColor));
  BC=XPal[BGColor];
  if(ScreenON) 
    {
      BorderChange(Y,BC,BC,240,9);
      p= EVEN_ODD&&(INTLAC||eo_flag); 
      S=(UseFont&&FontBuf)? FontBuf:VRAM+CHRGEN(0,0x800);
      S+=((Y+VDP[23])&0x07); 

      if (postproc) { 
	PixLine0[PL_ind](p,Y,S,linbuf); 
	postproc(ScrPtr(Y),240); 
      } 
      else PixLine0[PL_ind](p,Y,S,ScrPtr(Y));
      
      if (high) 
	{
	  if(EVEN_ODD && INTLAC) 
	    {
	      if (postproc) { 
		PixLine0[PL_ind](0,Y,S,linbuf); 
		postproc(ScrPtr_Odd(Y),240); 
	      } 
	      else PixLine0[PL_ind](0,Y,S,ScrPtr_Odd(Y));
	    } 
	  else NoIntLac(Y);
	}
    }
  else BorderChange(Y,BC,BC,0,0);
}

/*  SCREEN 1: TXT 32x24/26.5
 *  VDP[23] scrolls 256 screen-lines (32 textlines), VDP[26] is active
 *  Table usage (for area scrolled by VDP[23]):
 *    name    table (chrtab, len=0x400, xmsk=0x17FFF (?) ) : 0x400 bytes 
 *    colour  table (coltab, len= 0x40, xmsk=0x0FFFF (?) ) : 0x20 bytes
 *    pattern table (chrgen, len=0x800, xmsk=0x1FFFF (?) ) : 0x800 bytes
 *  Sprites: monochrome (see MonoSprites())
 */
# define PixLine1SprX(type,X,dest,inc)			       		 \
static void 								 \
PixLine1Spr##X##_##type(int p,int base,byte *S,byte *ColTab,void *Buf)   \
{									 \
  byte *SP, *T;								 \
  int lx,X,J,K;								 \
  type fc,bc;								 \
  type *buf;								 \
									 \
  buf=(type *)Buf;							 \
  SP=SBuf-1;								 \
  lx=VDP[26];								 \
  T=(VRAM+CHRTAB(base+(lx&0x1f),p || ((SP2) && ((lx&0x20)==0)),0x400));	 \
  for ( X=32 ; X ; X--,T++,NEXTCHAR(lx,T,base,p,0x400))			 \
    {									 \
      K=*( S+( (long)*T<<3 ) ); J=ColTab[(*T)>>3];		       	 \
      bc=XPal[J&0x0f].ct_##type; fc=XPal[J>>4].ct_##type;		 \
      for ( J=0x80 ; J ; J>>=1,inc)				       	 \
	dest=((*++SP) ? XPal[(*SP)&0x0F].ct_##type : (K&J)?fc:bc);	 \
    }									 \
}
#define PixLine1NoSprX(type,X,dest,inc)		       			    \
static void 								    \
PixLine1NoSpr##X##_##type(int p,int base,byte *S,byte * ColTab,void *Buf)   \
{									    \
  byte *T;								    \
  int lx,X,J,K;								    \
  type fc,bc;								    \
  type *buf;								    \
									    \
  buf=(type *)Buf;							    \
  lx=VDP[26];								    \
  T=(VRAM+CHRTAB(base+(lx&0x1f),p || ((SP2) && ((lx&0x20)==0)),0x400));	    \
  for ( X=32 ; X ; X--,T++,NEXTCHAR(lx,T,base,p,0x400))			    \
    {									    \
      K=*( S+( (long)*T<<3 ) ); J=ColTab[(*T)>>3];		       	    \
      bc=XPal[J&0x0f].ct_##type; fc=XPal[J>>4].ct_##type;		    \
      for ( J=0x80 ; J ; J>>=1,inc)  dest=((K&J)?fc:bc);		    \
    }									    \
}

XBLOATIFY(PixLine1Spr,  (int p, int base, byte *S, byte * ColTab, void *Buf))
XBLOATIFY(PixLine1NoSpr,(int p, int base, byte *S, byte * ColTab, void *Buf))

void RefreshLine1(register byte Y)
{
  register byte V,spr;
  int lx,p;
  ColTyp XPal0;
  register byte *S;

  if (!Mapped) return; 
  INIPAL(0xffff);
  if(ScreenON)
    {
      XPal0=XPal[0];
      if (! SolidColor0) XPal[0]=XPal[BGColor];
      BorderChange(Y,XPal[BGColor],XPal[BGColor],256,0);
      lx= VDP[26] & 0x3f;
      p = (EVEN_ODD&&(INTLAC||eo_flag));
      V=Y+VDP[23];
      S=(UseFont&&FontBuf) ? FontBuf : VRAM+CHRGEN(0,0x800); 
      S+=(V&0x07); 
      spr=(!SpritesOFF)&&MonoSprites(V);
      if (postproc) {
	(spr?PixLine1Spr:PixLine1NoSpr)[PL_ind]
	  (p, (V>>3)<<5, S, VRAM+COLTAB(0,0x40),linbuf);
	postproc(ScrPtr(Y),256);
      }
      else (spr?PixLine1Spr:PixLine1NoSpr)[PL_ind]
	     (p, (V>>3)<<5, S, VRAM+COLTAB(0,0x40),ScrPtr(Y));
      if (high) 
	{
	  if (INTERLACING) 
	    {
	      if (postproc) {
		(spr?PixLine1Spr:PixLine1NoSpr)[PL_ind]
		  (0, (V>>3)<<5, S, VRAM+COLTAB(0,0x40),linbuf);
		postproc(ScrPtr_Odd(Y),256);
	      }
	      else (spr?PixLine1Spr:PixLine1NoSpr)[PL_ind]
		     (0, (V>>3)<<5, S, VRAM+COLTAB(0,0x40),ScrPtr_Odd(Y));
	    }
	  else NoIntLac(Y);
	}
      XPal[0]=XPal0;
    }
  else BorderChange(Y,XPal[BGColor],XPal[BGColor],0,0);
}

/*  SCREEN 2,4: BLK 256x192/212
 *  VDP[23] scrolls 256 screen-lines, VDP[26] is active
 *  Table usage (for area scrolled by VDP[23]):
 *    name    table (chrtab, stp=len=siz=0x400, xmsk=0x17FFF (?) )
 *    colour  table (coltab, stp= 0x40, len=siz=0x2000, xmsk=0x0FFFF (?) )
 *    pattern table (chrgen, stp=0x800, len=siz=0x2000, xmsk=0x1FFFF (?) )
 *  Sprites: 
 *    - SCREEN 2: monochrome (see MonoSprites())
 *    - SCREEN 4: coloured (see ColorSprites())
 */
#define PixLine2SprX(type,X,dest,inc)		      			\
static void 								\
PixLine2Spr##X##_##type(int p,int base,byte * PGT,int CLT,void *Buf)	\
{									\
  byte *SP,*T, K;							\
  int lx,X,i;								\
  type fc,bc;								\
  type *buf;								\
									\
  buf=(type*)Buf;							\
  SP=SBuf-1;								\
  lx= VDP[26];								\
  T=(VRAM+CHRTAB(base+(lx&0x1f),p || ((SP2) && ((lx&0x20)==0)),0x400));	\
  for ( X=32 ; X ; X--,T++,NEXTCHAR(lx,T,base,p,0x400))			\
    {									\
      K=PGT[(*T)<<3]; 							\
      i=VRAM[COLTAB(CLT+((*T)<<3),0x2000)];				\
      fc=XPal[i>>4].ct_##type; bc=XPal[i&0x0F].ct_##type;		\
      for (i=0x80 ; i ; i>>=1,inc) 	       				\
	dest= (*++SP) ? XPal[*SP&0x0F].ct_##type			\
	              :((K&i)?fc:bc);					\
    }									\
}
#define PixLine2NoSprX(type,X,dest,inc)		       			\
static void 								\
PixLine2NoSpr##X##_##type(int p,int base,byte * PGT,int CLT,void *Buf)	\
{									\
  byte *T, K;								\
  int lx,X,i;								\
  type fc,bc;								\
  type *buf;								\
									\
  buf=(type*)Buf;							\
  lx= VDP[26];								\
  T=(VRAM+CHRTAB(base+(lx&0x1f),p || ((SP2) && ((lx&0x20)==0)),0x400));	\
  for ( X=32 ; X ; X--,T++,NEXTCHAR(lx,T,base,p,0x400))			\
    {									\
      K=PGT[(*T)<<3]; 							\
      i=VRAM[COLTAB(CLT+((*T)<<3),0x2000)];				\
      fc=(XPal[i>>4].ct_##type); bc=(XPal[i&0x0F].ct_##type);      	\
      for (i=0x80 ; i ; i>>=1, inc)  dest=((K&i)?fc:bc);		\
    }									\
}

XBLOATIFY(PixLine2Spr,  (int p, int base, byte *S, int, void *Buf))
XBLOATIFY(PixLine2NoSpr,(int p, int base, byte *S, int, void *Buf))

#define REFRESH_LINE_BLK(n,sprmod)				\
void RefreshLine##n(register byte Y)				\
{								\
  register byte V, spr;						\
  ColTyp XPal0;							\
  int p;							\
  register int CLT;						\
  register byte *PGT;						\
  								\
  if (!Mapped) return;						\
  INIPAL(0xffff);						\
  if(ScreenON)							\
    {								\
      XPal0=XPal[0];						\
      if (! SolidColor0) XPal[0]=XPal[BGColor];			\
      BorderChange(Y,XPal[BGColor],XPal[BGColor],256,0);	\
      p= (EVEN_ODD&&(INTLAC||eo_flag)); 			\
      V=Y+VDP[23];						\
      CLT=((int)(V>>6)<<11)+(V&0x07);				\
      PGT=VRAM+CHRGEN(CLT, 0x2000);				\
      spr=(!SpritesOFF)&&sprmod##Sprites(V);			\
      if (postproc) {						\
	(spr?PixLine2Spr:PixLine2NoSpr)[PL_ind]			\
	  (p,(V>>3)<<5,PGT,CLT,linbuf);				\
	postproc(ScrPtr(Y),256);				\
      }								\
      else  (spr?PixLine2Spr:PixLine2NoSpr)[PL_ind]		\
	      (p,(V>>3)<<5,PGT,CLT,ScrPtr(Y));			\
      if (high) {						\
	if (INTERLACING) {					\
	  if (postproc) {					\
	    (spr?PixLine2Spr:PixLine2NoSpr)[PL_ind]		\
	      (0,(V>>3)<<5,PGT,CLT,linbuf);			\
	    postproc(ScrPtr_Odd(Y),256);			\
	  }							\
	  else  (spr?PixLine2Spr:PixLine2NoSpr)[PL_ind]		\
		  (0,(V>>3)<<5,PGT,CLT,ScrPtr_Odd(Y));		\
	}							\
	else NoIntLac(Y);					\
      }								\
      XPal[0]=XPal0;						\
    }								\
  else BorderChange(Y,XPal[BGColor],XPal[BGColor],0,0);		\
}

REFRESH_LINE_BLK(2,Mono)
REFRESH_LINE_BLK(4,Color)

/*  SCREEN 3: BLK 64x48/53x16
 *  VDP[23] scrolls 256 screen-lines (64 rows of pixelblocks)
 *  Table usage (for area scrolled by VDP[23]):
 *    name    table (chrtab, stp=len=siz=0x400, xmsk=0x????? )
 *    pattern table (chrgen, stp=len=siz=0x800, xmsk=0x????? )
 *  Sprites: monochrome (see MonoSprites())
 */
#define PixLine3SprX(type,X,dest,inc)			       		\
static void 								\
PixLine3Spr##X##_##type(int p, int base, byte * CG, byte V, void * Buf)	\
{									\
  byte *SP,*T;								\
  int lx,X;								\
  type *buf;								\
									\
  buf=(type*)Buf;							\
  SP=SBuf-1;								\
  lx= VDP[26];								\
  T=(VRAM+CHRTAB(base+(lx&0x1f),p || ((SP2) && ((lx&0x20)==0)),0x400));	\
  for ( X=0 ; X<32 ; X++, T++,NEXTCHAR(lx,T,base,p,0x400))		\
    {									\
      register byte C,J;						\
      type C1,C2;							\
      									\
      C=CG[((int)*T<<3)+((V>>2)&0x07)];					\
      C1=XPal[C>>4].ct_##type; C2=XPal[C&0x0F].ct_##type;		\
      for ( J=4 ; J ; J--, inc)						\
	dest = *(++SP) ? XPal[(*SP)&0x0F].ct_##type : C1;		\
      for ( J=4 ; J ; J--, inc)						\
	dest = *(++SP) ? XPal[(*SP)&0x0F].ct_##type : C2;		\
    }									\
}
#define PixLine3NoSprX(type,X,dest,inc)			       		\
static void 								\
PixLine3NoSpr##X##_##type(int p,int base,byte * CG,byte V,void *Buf)   	\
{									\
  byte *T;								\
  int lx,X;								\
  type *buf;								\
									\
  buf=(type*) Buf;							\
  lx= VDP[26];								\
  T=(VRAM+CHRTAB(base+(lx&0x1f),p || ((SP2) && ((lx&0x20)==0)),0x400));	\
  for ( X=0 ; X<32 ; X++, T++,NEXTCHAR(lx,T,base,p,0x400))		\
    {									\
      register byte C;							\
	  								\
      C=CG[((int)*T<<3)+((V>>2)&0x07)];					\
      dest=XPal[C>>4].ct_##type; inc;					\
      dest=XPal[C&0x0f].ct_##type; inc;					\
    }									\
}

#define PixLine3NoSprN(type) PixLine3NoSprX(type,N,			\
  buf[0]=buf[1]=buf[2]=buf[3],buf+=4)

#define PixLine3NoSprW(type) PixLine3NoSprX(type,W,			\
  buf[0]=buf[1]=buf[2]=buf[3]=buf[4]=buf[5]=buf[6]=buf[7],buf+=8)

XBLOATIFY(PixLine3Spr,(int p, int base, byte * CG, byte V,void *))
BLOATIFY(PixLine3NoSpr,(int p, int base, byte * CG, byte V,void *))

void RefreshLine3(register byte Y)
{
  register byte V;
  byte spr;
  int lx,p;
  register byte *CG;
  ColTyp XPal0;

  if (!Mapped) return;
  INIPAL(0xffff);
  if (ScreenON)
    {
      XPal0=XPal[0];
      if (!SolidColor0) XPal[0]=XPal[BGColor];
      BorderChange(Y,XPal[BGColor],XPal[BGColor],256,0);
      lx= VDP[26] & 0x3f;
      p=(EVEN_ODD&&(INTLAC||eo_flag));
      V=Y+VDP[23];

      CG=VRAM+CHRGEN(0,0x800);
      spr=(!SpritesOFF)&&MonoSprites(V);
      if (postproc) {
	(spr?PixLine3Spr:PixLine3NoSpr)[PL_ind]
	  (p,(V>>3)<<5,CG,V,linbuf);
	postproc(ScrPtr(Y),256);
      }
      else (spr?PixLine3Spr:PixLine3NoSpr)[PL_ind]
	     (p,(V>>3)<<5,CG,V,ScrPtr(Y));
      
      if (high) 
       {
	 if (INTERLACING)  /* interlacing in screen 3, ridiculous! */
	   {
	     if (postproc) {
	       (spr?PixLine3Spr:PixLine3NoSpr)[PL_ind]
		 (0,(V>>3)<<5,CG,V,linbuf);
	       postproc(ScrPtr_Odd(Y),256);
	     }
	     else (spr?PixLine3Spr:PixLine3NoSpr)[PL_ind]
		    (0,(V>>3)<<5,CG,V,ScrPtr_Odd(Y));
	   }
	 else NoIntLac(Y);
       }
     XPal[0]=XPal0;
    }
  else BorderChange(Y,XPal[BGColor],XPal[BGColor],0,0);
}

/*  SCREEN 4: see SCREEN 2,4 above */

/*  SCREEN 5: bitmap 256x192/212x16
 *  VDP[23] scrolls 256 screen-lines
 *  Table usage (for area scrolled by VDP[23]):
 *    matrix  table (chrtab, stp=0x400, len=siz=0x8000, xmsk=0x17fff )
 *  Sprites: Coloured (see ColorSprites)
 */
#define PixLine5SprX(type,X,dest,inc)					    \
static void PixLine5Spr##X##_##type(int p, int base, void *Buf)		    \
{									    \
  byte *T, *SP;								    \
  int lx,X;								    \
  type *buf;								    \
									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];							       	    \
  T=VRAM+CHRTAB(base+4*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x8000);    \
  SP=SBuf-1;								    \
  for ( X=32 ; X ; X--,NEXTCHAR(lx,T,base,p,0x8000))			    \
    {									    \
      int i;								    \
      for (i=4 ; i ; i--,T++) 						    \
	{								    \
	  dest=((*++SP)?(XPal[(*SP)&0x0F]):(XPal[(*T)>>4])).ct_##type; 	    \
	  inc;								    \
	  dest=((*++SP)?(XPal[(*SP)&0x0F]):(XPal[(*T)&0x0F])).ct_##type;    \
	  inc;								    \
	}								    \
      } 								    \
}
#define PixLine5NoSprX(type,X,dest,inc)					 \
static void PixLine5NoSpr##X##_##type(int p, int base, void * Buf)	 \
{									 \
  byte *T;								 \
  int lx,X;								 \
  type *buf;								 \
									 \
  buf=(type*)Buf;							 \
  lx= VDP[26];								 \
  T=VRAM+CHRTAB(base+4*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x8000); \
  for ( X=32 ; X ; X--,NEXTCHAR(lx,T,base,p,0x8000))			 \
    {									 \
      dest=XPal[(*T)>>4].ct_##type; inc;				 \
      dest=XPal[*(T++)&0x0f].ct_##type; inc;				 \
      dest=XPal[(*T)>>4].ct_##type; inc;				 \
      dest=XPal[*(T++)&0x0f].ct_##type; inc;				 \
      dest=XPal[(*T)>>4].ct_##type; inc;				 \
      dest=XPal[*(T++)&0x0f].ct_##type; inc;				 \
      dest=XPal[(*T)>>4].ct_##type; inc;				 \
      dest=XPal[*(T++)&0x0f].ct_##type; inc;				 \
    }									 \
} 

XBLOATIFY(PixLine5Spr,(int p, int base, void * Buf))
XBLOATIFY(PixLine5NoSpr,(int p, int base, void * Buf))
 
void RefreshLine5(register byte Y)
{
  register byte V;
  int lx,p;
  register byte spr;
  ColTyp XPal0;
  
  if (!Mapped) return;
  INIPAL(0xffff);
  if (ScreenON)
    {
      XPal0=XPal[0];
      if (!SolidColor0) XPal[0]=XPal[BGColor];
      BorderChange(Y,XPal[BGColor],XPal[BGColor],256,0);
      lx= VDP[26] & 0x3f;
      p= (EVEN_ODD&&(INTLAC||eo_flag));
      V=Y+VDP[23];
 
      spr = (!SpritesOFF)&&ColorSprites(V);
      if (postproc) {
	((spr?PixLine5Spr:PixLine5NoSpr)[PL_ind])(p, 128*V, linbuf);
	postproc(ScrPtr(Y),256);
      }
      else 
	((spr?PixLine5Spr:PixLine5NoSpr)[PL_ind])(p, 128*V, ScrPtr(Y));
      if (high) {
	if (INTERLACING) {
	  if (postproc) {
	    ((spr?PixLine5Spr:PixLine5NoSpr)[PL_ind])(0, 128*V, linbuf);
	    postproc(ScrPtr_Odd(Y),256);
	  }
	  else 
	    ((spr?PixLine5Spr:PixLine5NoSpr)[PL_ind])(0,128*V,ScrPtr_Odd(Y));
	}
	else NoIntLac(Y);
      }
      XPal[0]=XPal0; 
    }
  else BorderChange(Y,XPal[BGColor],XPal[BGColor],0,0);
}

/*  SCREEN 6: bitmap 512x192/212x4
 *  VDP[23] scrolls 256 screen-lines
 *  Table usage (for area scrolled by VDP[23]):
 *    matrix  table (chrtab, stp=0x400, len=siz=0x8000, xmsk=0x17fff )
 *  Sprites: Coloured (see ColorSprites)
 */
#define PixLine6SprW(type)						 \
static void 								 \
PixLine6SprW_##type(int p, int base, ColTyp RPal[], void *Buf)		 \
{									 \
  byte *T, *SP;								 \
  int lx,X;								 \
  type *buf;								 \
									 \
  buf=(type*)Buf;							 \
  lx= VDP[26];								 \
  T=VRAM+CHRTAB(base+4*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x8000); \
  SP=SBuf-1;								 \
  for ( X=32 ; X ; X--,NEXTCHAR(lx,T,base,p,0x8000))			 \
    {									 \
      int i;								 \
      for (i=4 ; i ; --i, T++) {					 \
	if (*++SP) 							 \
	  { *(buf++)=XPal[((*SP)>>2)&0x03].ct_##type;			 \
	    *(buf++)=RPal[(*SP)&0x03].ct_##type; }			 \
	else								 \
	  { *(buf++)=XPal[*(T)>>6].ct_##type;				 \
	    *(buf++)=RPal[(*(T)>>4)&0x03].ct_##type; }			 \
	if (*++SP) 							 \
	  { *(buf++)=XPal[((*SP)>>2)&0x03].ct_##type;			 \
	    *(buf++)=RPal[(*SP)&0x03].ct_##type; }			 \
	else								 \
	  { *(buf++)=XPal[(*(T)>>2)&0x03].ct_##type;			 \
	    *(buf++)=RPal[(*(T))&0x03].ct_##type; } 			 \
      }									 \
    }									 \
}
#define PixLine6NoSprW(type)						     \
static void PixLine6NoSprW_##type(int p, int base, ColTyp RPal[], void *Buf) \
{									     \
  byte *T;								     \
  int lx,X;								     \
  type *buf;								     \
									     \
  buf=(type*)Buf;							     \
  lx= VDP[26];								     \
  T=VRAM+CHRTAB(base+4*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x8000);     \
  for ( X=32 ; X ; X--,NEXTCHAR(lx,T,base,p,0x8000))			     \
    {									     \
      int i;								     \
									     \
      for(i=4 ; i ; i--, T++)						     \
	{ 								     \
	  *(buf++)=XPal[*(T)>>6].ct_##type;				     \
	  *(buf++)=RPal[(*(T)>>4)&0x03].ct_##type;			     \
	  *(buf++)=XPal[(*(T)>>2)&0x03].ct_##type;			     \
	  *(buf++)=RPal[(*(T))&0x03].ct_##type;				     \
	}								     \
    }									     \
}
#define PixLine6SprN(type)						   \
static void PixLine6SprN_##type(int p, int base, ColTyp RPal[], void *Buf) \
{									   \
  byte *T, *SP;								   \
  int lx,X;								   \
  type *buf;								   \
									   \
  buf=(type*)Buf;							   \
  lx= VDP[26];								   \
  T=VRAM+CHRTAB(base+4*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x8000);   \
  SP=SBuf-1;								   \
  for ( X=32 ; X ; X--,NEXTCHAR(lx,T,base,p,0x8000))			   \
    {									   \
      int i;								   \
      for (i=4 ; i ; --i, T++) {					   \
	if (*++SP) 							   \
	  *(buf++)=XPal[((*SP)>>2)&0x03].ct_##type;			   \
	else								   \
	  *(buf++)=XPal[*(T)>>6].ct_##type;				   \
	if (*++SP) 							   \
	  *(buf++)=XPal[((*SP)>>2)&0x03].ct_##type;			   \
	else								   \
	  *(buf++)=XPal[(*(T)>>2)&0x03].ct_##type;			   \
      }									   \
    }									   \
}
#define PixLine6NoSprN(type)						     \
static void PixLine6NoSprN_##type(int p, int base, ColTyp RPal[], void *Buf) \
{									     \
  byte *T;								     \
  int lx,X;								     \
  type *buf;								     \
									     \
  buf=(type*)Buf;							     \
  lx= VDP[26];								     \
  T=VRAM+CHRTAB(base+4*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x8000);     \
  for ( X=32 ; X ; X--,NEXTCHAR(lx,T,base,p,0x8000))			     \
    {									     \
      int i;								     \
									     \
      for(i=4 ; i ; i--, T++)						     \
	{ 								     \
	  *(buf++)=XPal[*(T)>>6].ct_##type;				     \
	  *(buf++)=XPal[(*(T)>>2)&0x03].ct_##type;			     \
	}								     \
    }									     \
}

BLOATIFY(PixLine6Spr,(int p, int base, ColTyp RPal[], void *Buf))
BLOATIFY(PixLine6NoSpr,(int p, int base, ColTyp RPal[], void *Buf))

void RefreshLine6(register byte Y)
{
  register byte V;
  int p,spr;
  ColTyp XPal0;
  static ColTyp RPal[4];
  
  if (!Mapped) return;
  INIPAL(0x000f);
  if (ScreenON)
    {
      XPal0=XPal[0];
      if (!SolidColor0)
	{ 
	  XPal[0]=XPal[BGColor>>2];
	  RPal[0]=XPal[BGColor&3];
	  memcpy(RPal+1, XPal+1, 3*sizeof(ColTyp));
	} 
      else 
	memcpy(RPal, XPal, 4*sizeof(ColTyp));
      BorderChange(Y,XPal[BGColor>>2],XPal[BGColor&3],256,0);
      V=Y+VDP[23];
      p=(EVEN_ODD&&(INTLAC||eo_flag));
      spr=(!SpritesOFF)&&ColorSprites(V);
      if (postproc) {
	((spr?PixLine6Spr:PixLine6NoSpr)[PL_ind])(p, 128*V, RPal, linbuf);
	postproc(ScrPtr(Y),256);
      }
      else 
	((spr?PixLine6Spr:PixLine6NoSpr)[PL_ind])(p, 128*V, RPal, ScrPtr(Y));
      if (high) {
	if (INTERLACING) {
	  if (postproc) {
	    ((spr?PixLine6Spr:PixLine6NoSpr)[PL_ind])(0, 128*V, RPal, linbuf);
	    postproc(ScrPtr_Odd(Y),256);
	  }
	  else 
	    ((spr?PixLine6Spr:PixLine6NoSpr)[PL_ind])(0,128*V,RPal,
						      ScrPtr_Odd(Y));
	}
	else NoIntLac(Y);
      }
      XPal[0]=XPal0;
    }
  else BorderChange(Y,XPal[BGColor>>2],XPal[BGColor&3],0,0);
}

/*  SCREEN 7: bitmap 512x192/212x16
 *  VDP[23] scrolls 256 screen-lines 
 *  Table usage (for area scrolled by VDP[23]):
 *    matrix  table (chrtab78, stp=0x800,len=siz=0x10000, xmsk=0x0ffff)
 *  Sprites: Coloured (see ColorSprites)
 */
#define PixLine7SprW(type)						    \
static void PixLine7SprW_##type(int p, int base, void *Buf)		    \
{									    \
  byte *T, *SP;								    \
  int lx,X;								    \
  type *buf;								    \
									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];								    \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000); \
  SP=SBuf-1;								    \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			    \
    {									    \
      int i;								    \
      for (i=8 ; i ; i--,T++) { 					    \
	if (*++SP) 							    \
	  { 								    \
	    buf[0]=buf[1]=XPal[(*SP)&0x0F].ct_##type;	 		    \
	    buf+=2;							    \
	  } 								    \
	else 								    \
	  {								    \
	    *(buf++)=XPal[*(T)>>4].ct_##type;				    \
	    *(buf++)=XPal[*(T)&0x0F].ct_##type;				    \
	  }								    \
      }									    \
    }									    \
} 
#define PixLine7NoSprW(type)						    \
static void PixLine7NoSprW_##type(int p, int base, void * Buf)		    \
{									    \
  byte *T;								    \
  int lx,X;								    \
  type *buf;								    \
									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];								    \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000); \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			    \
    {									    \
      int i;								    \
      for (i=8 ; i ; i--,T++)						    \
	{								    \
	  *(buf++)=XPal[*(T)>>4].ct_##type;				    \
	  *(buf++)=XPal[*(T)&0x0F].ct_##type;			    	    \
	} 								    \
    }									    \
} 
#define PixLine7SprN(type)						    \
static void PixLine7SprN_##type(int p, int base, void *Buf)		    \
{									    \
  byte *T, *SP;								    \
  int lx,X;								    \
  type *buf;								    \
									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];								    \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000); \
  SP=SBuf-1;								    \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			    \
    {									    \
      int i;								    \
      for (i=8 ; i ; i--,T++) 					    \
	*(buf++) = ((*++SP)  						    \
		    ? XPal[(*SP)&0x0F] 		                            \
	    	    : (XPal[*(T)>>4])).ct_##type;			    \
    }									    \
} 
#define PixLine7NoSprN(type)						    \
static void PixLine7NoSprN_##type(int p, int base, void * Buf)		    \
{									    \
  byte *T;								    \
  int lx,X;								    \
  type *buf;								    \
									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];								    \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000); \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			    \
    {									    \
      int i;								    \
      for (i=8 ; i ; i--,T++)						    \
	  *(buf++)=XPal[*(T)>>4].ct_##type;				    \
    }									    \
} 

BLOATIFY(PixLine7Spr,(int p, int base, void * Buf))
BLOATIFY(PixLine7NoSpr,(int p, int base, void * Buf))

void RefreshLine7(register byte Y)
{
  register byte V;
  int p,spr;
  ColTyp XPal0;

  if (!Mapped) return;
  INIPAL(0xffff);
  if (ScreenON)
    {
      XPal0=XPal[0];
      if (!SolidColor0) XPal[0]=XPal[BGColor];
      BorderChange(Y,XPal[BGColor],XPal[BGColor],256,0);
      V=Y+VDP[23];
      p= (EVEN_ODD&&(INTLAC||eo_flag));
      spr = (!SpritesOFF)&&ColorSprites(V);
      if (postproc) {
	((spr?PixLine7Spr:PixLine7NoSpr)[PL_ind])(p, 256*V, linbuf);
	postproc(ScrPtr(Y),256);
      }
      else 
	((spr?PixLine7Spr:PixLine7NoSpr)[PL_ind])(p, 256*V, ScrPtr(Y));
      if (high) {
	if (INTERLACING) {
	  if (postproc) {
	    ((spr?PixLine7Spr:PixLine7NoSpr)[PL_ind])(0, 256*V, linbuf);
	    postproc(ScrPtr_Odd(Y),256);
	  }
	  else 
	    ((spr?PixLine7Spr:PixLine7NoSpr)[PL_ind])(0,256*V,ScrPtr_Odd(Y));
	}
	else NoIntLac(Y);
      } 
      XPal[0]=XPal0; 
    }
  else BorderChange(Y,XPal[BGColor],XPal[BGColor],0,0);
}


/*  SCREEN 8: bitmap 256x192/212x256
 *  VDP[23] scrolls 256 screen-lines
 *  Table usage (for area scrolled by VDP[23]):
 *    matrix  table (chrtab78, stp=0x800,len=siz=0x10000, xmsk=0x0ffff)
 *  Sprites: Coloured (see ColorSprites)
 *    note: the sprite colours are fixed (
 */
#define PixLine8SprX(type,X,dest,inc)					    \
static void PixLine8Spr##X##_##type(int p, int base, void * Buf)       	    \
{									    \
  byte *T, *SP;								    \
  int lx,X;								    \
  type *buf;								    \
									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];								    \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000); \
  SP=SBuf-1;								    \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			    \
    {									    \
      int i;								    \
      for (i=8; i ; i--, T++, inc)					    \
	dest=(XPal8[(*SP)?SprCol8[(*SP)&0x0F]:*(T)]).ct_##type;	    	    \
    }									    \
}
#define PixLine8NoSprX(type,X,dest,inc) 				     \
static void PixLine8NoSpr##X##_##type(int p, int base, void * Buf)	     \
{									     \
  byte *T;								     \
  int lx,X;								     \
  type *buf;								     \
									     \
  buf=(type*)Buf;							     \
  lx= VDP[26];								     \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0) ), 0x10000); \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			     \
    {									     \
      int i;								     \
      for (i=8 ; i ; i--,T++,inc)					     \
	dest=(XPal8[*(T)]).ct_##type;					     \
    }									     \
} 

XBLOATIFY(PixLine8Spr,(int p, int base, void * Buf))
XBLOATIFY(PixLine8NoSpr,(int p, int base, void * Buf))

void RefreshLine8(register byte Y)
{
  register byte V;
  int spr,p;
  /* ColTyp XPal0; not needed, there is no transparent color in screen 8 */

  if (!Mapped) return;
  if (!IsPal8) Colors.TakeScr8();  /* overkill if (! ScreenON) */
  Used8=1;
  if (ScreenON)
    {
      V=Y+VDP[23];
      BorderChange(Y,XPal8[VDP[7]],XPal8[VDP[7]],256, 0);
      p=EVEN_ODD&&(INTLAC||eo_flag);

      spr=(!SpritesOFF)&&ColorSprites(V);
      if (postproc) {
	((spr?PixLine8Spr:PixLine8NoSpr)[PL_ind])(p, 256*V, linbuf);
	postproc(ScrPtr(Y),256);
      }
      else 
	((spr?PixLine8Spr:PixLine8NoSpr)[PL_ind])(p, 256*V, ScrPtr(Y));
      if (high) {
	if (INTERLACING) {
	  if (postproc) {
	    ((spr?PixLine8Spr:PixLine8NoSpr)[PL_ind])(0, 256*V, linbuf);
	    postproc(ScrPtr_Odd(Y),256);
	  }
	  else 
	    ((spr?PixLine8Spr:PixLine8NoSpr)[PL_ind])(0,256*V,ScrPtr_Odd(Y));
	}
	else NoIntLac(Y);
      }
    }
  else  BorderChange(Y, XPal8[VDP[7]], XPal8[VDP[7]], 0,0);
}


/*  SCREEN 10/11: 256x192/212x(YJK+16)
 *  VDP[23] scrolls 256 screen-lines
 *  Table usage (for area scrolled by VDP[23]):
 *    matrix  table (chrtab78, stp=0x800,len=siz=0x10000, xmsk=0x0ffff)
 *  Sprites: Coloured (see ColorSprites)
 */
#define PixLineASprX(type,X,dest,inc)					      \
static void PixLineASpr##X##_##type(int p, int base, void * Buf)	      \
{									      \
  byte *SP, *T;								      \
  int lx,X;								      \
  type *buf;								      \
									      \
  buf=(type*)Buf;							      \
  lx= VDP[26];								      \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000);   \
  SP=SBuf-1;								      \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			      \
    {									      \
      int i,Y, J,K;							      \
									      \
      getJK(T, &J, &K);							      \
      for (i=4 ; i ; i--, T++,inc)					      \
	if (*++SP)							      \
	  dest=XPal[*SP&0x0F].ct_##type;				      \
	else 								      \
	  { Y=*(T)>>3; dest=(Y&1 ? XPal[Y>>1] : YJKcolor(Y,J,K)).ct_##type; } \
      getJK(T, &J, &K);							      \
      for (i=4 ; i ; i--, T++,inc)					      \
	if (*++SP)							      \
	  dest=XPal[*SP&0x0F].ct_##type;				      \
	else 								      \
	  { Y=*(T)>>3; dest=(Y&1 ? XPal[Y>>1] : YJKcolor(Y,J,K)).ct_##type; } \
    }									      \
}
#define PixLineANoSprX(type,X,dest,inc)					    \
static void PixLineANoSpr##X##_##type(int p, int base, void * Buf)	    \
{									    \
  byte *T;								    \
  int lx,X;								    \
  type *buf;								    \
									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];								    \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000); \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			    \
    {									    \
      int Y,J,K,i;							    \
									    \
      getJK(T, &J, &K);							    \
      for (i=4; i ; i--, inc)						    \
	{								    \
	  Y=*(T++)>>3; 							    \
	  dest=( Y&1 ? XPal[Y>>1] : YJKcolor(Y,J,K) ).ct_##type; 	    \
	}								    \
      getJK(T, &J, &K);							    \
      for (i=4; i ; i--, inc)						    \
	{								    \
	  Y=*(T++)>>3; 							    \
	  dest=( Y&1 ? XPal[Y>>1] : YJKcolor(Y,J,K) ).ct_##type; 	    \
	}								    \
    }									    \
}

XBLOATIFY(PixLineASpr,(int p, int base, void * Buf))
XBLOATIFY(PixLineANoSpr,(int p, int base, void * Buf))

void
RefreshLine10(register byte Y)
{
  int spr, pg;
  ColTyp XPal0;
  byte V;

  if (!Mapped) return;
  if (!IsPal8) Colors.TakeScr8();  /* overkill if (! ScreenON) */
  Used8=1;
  if (ScreenON)
    {
      BorderChange(Y, XPal[BGColor], XPal[BGColor], 256,4);
      V=Y+VDP[23];
      INIPAL(0xFFFF);
      XPal0=XPal[0];
      if (! SolidColor0)  XPal[0]=XPal[BGColor];
      pg=EVEN_ODD&&(INTLAC||eo_flag);
      spr=(!SpritesOFF)&&ColorSprites(V);
      if (postproc) {
	((spr?PixLineASpr:PixLineANoSpr)[PL_ind])(pg, 256*V, linbuf);
	postproc(ScrPtr(Y),256);
      }
      else 
	((spr?PixLineASpr:PixLineANoSpr)[PL_ind])(pg, 256*V, ScrPtr(Y));
      if (high) {
	if (INTERLACING) {
	  if (postproc) {
	    ((spr?PixLineASpr:PixLineANoSpr)[PL_ind])(0, 256*V, linbuf);
	    postproc(ScrPtr_Odd(Y),256);
	  }
	  else 
	    ((spr?PixLineASpr:PixLineANoSpr)[PL_ind])(0, 256*V, ScrPtr_Odd(Y));
	}
	else NoIntLac((V-VDP[23])&0xFF);
      }
      XPal[0]=XPal0;
    }
  else  BorderChange(Y, XPal[BGColor],XPal[BGColor], 0,0);
}

/*  SCREEN 12: bitmap 256x192/212x(YJK)
 *  VDP[23] scrolls 256 screen-lines
 *  Table usage (for area scrolled by VDP[23]):
 *    matrix  table (chrtab78, stp=0x800,len=siz=0x10000, xmsk=0x0ffff)
 *  Sprites: Coloured (see ColorSprites)
 */
#define PixLineCSprX(type,X,dest,inc)					    \
static void PixLineCSpr##X##_##type(int p, int base, void * Buf)	    \
{									    \
  byte *SP, *T;								    \
  int lx,X,J,K,i;							    \
  type *buf;								    \
  									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];								    \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000); \
  SP=SBuf-1;								    \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			    \
    {									    \
      getJK(T, &J, &K);							    \
      for ( i=4 ; i ; i--,T++,inc)					    \
	dest=( *++SP ? XPal[*SP&0x0F] : YJKcolor(*(T)>>3,J,K)).ct_##type;   \
									    \
      getJK(T, &J, &K);							    \
      for ( i=4 ; i ; i--,T++,inc)					    \
	dest=( *++SP ? XPal[*SP&0x0F] : YJKcolor(*(T)>>3,J,K)).ct_##type;   \
    }									    \
}
#define PixLineCNoSprX(type,X,dest,inc)					    \
static void PixLineCNoSpr##X##_##type(int p, int base, void * Buf)     	    \
{									    \
  byte *T;								    \
  int lx,X,J,K,i;					       		    \
  type *buf;								    \
									    \
  buf=(type*)Buf;							    \
  lx= VDP[26];								    \
  T=VRAM+CHRTAB78(base+8*(lx&0x1f), p || (SP2 && ((lx&0x20)==0)), 0x10000); \
  for ( X=32 ; X ; X--,NEXTCHAR78(lx,T,base,p,0x10000))			    \
    {									    \
      getJK(T, &J, &K);							    \
      for ( i=4 ; i ; i--, T++, inc)					    \
	dest=( YJKcolor(*T>>3,J,K) ).ct_##type;				    \
      getJK(T, &J, &K);							    \
      for ( i=4 ; i ; i--, T++, inc)					    \
	dest=( YJKcolor(*T>>3,J,K) ).ct_##type;				    \
    }									    \
}

XBLOATIFY(PixLineCSpr,(int,int,void*));
XBLOATIFY(PixLineCNoSpr,(int,int,void*));

void
RefreshLine12(register byte Y)
{
  int spr,pg;
  ColTyp XPal0;
  byte V;

  if (!Mapped) return;
  if (!IsPal8) Colors.TakeScr8();  /* overkill if (! ScreenON) */
  Used8=1;
  if (ScreenON)
    {
      BorderChange(Y, XPal[BGColor], XPal[BGColor], 256,4);
      V=Y+VDP[23];
      INIPAL(0xFFFF);
      XPal0=XPal[0];
      if (! SolidColor0)  XPal[0]=XPal[BGColor];
      pg=EVEN_ODD&&(INTLAC||eo_flag);
      spr=(!SpritesOFF)&&ColorSprites(V);
      if (postproc) {
	((spr?PixLineCSpr:PixLineCNoSpr)[PL_ind])(pg, 256*V, linbuf);
	postproc(ScrPtr(Y),256);
      }
      else 
	((spr?PixLineCSpr:PixLineCNoSpr)[PL_ind])(pg, 256*V, ScrPtr(Y));
      if (high) {
	if (INTERLACING) {
	  if (postproc) {
	    ((spr?PixLineCSpr:PixLineCNoSpr)[PL_ind])(0, 256*V, linbuf);
	    postproc(ScrPtr_Odd(Y),256);
	  }
	  else 
	    ((spr?PixLineCSpr:PixLineCNoSpr)[PL_ind])(0, 256*V, ScrPtr_Odd(Y));
	}
	else NoIntLac((V-VDP[23])&0xFF);
      }
      XPal[0]=XPal0;
    }
  else  BorderChange(Y, XPal[BGColor], XPal[BGColor], 0,0);
}

/*  SCREEN 0: TXT 80x24/26.5
 *  VDP[23] scrolls blocks of 8 screen-lines (one textline) 
 *  Table usage:
 *    name    table (chrtab,stp=0x400,len=0x1000,siz=0x780/0x870,xmsk=0x17FFF)
 *    colour  table (coltab,stp= 0x40,len=0x200,siz=0xF0/0x10E,xmsk=0x0FFFF)
 *    pattern table (chrgen,stp=0x800,len=siz=0x800,xmsk=0x0FFFF)
 *  Sprites: none
 */
#define PixLineTx80W(type)					\
static void PixLineTx80W_##type(int p, byte Y, void *Buf)	\
{								\
  byte X,J,K,M;							\
  type bc,fc,xbc,xfc;						\
  byte *CG, *S;							\
  register long T, TMSK, C, CMSK;				\
  type *buf;							\
								\
  buf=(type*)Buf;						\
  bc=XPal[BGColor].ct_##type;					\
  fc=(FGColor||SolidColor0)?XPal[FGColor].ct_##type:bc;	\
  xbc=(XBGColor||SolidColor0)?XPal[XBGColor].ct_##type:bc;	\
  xfc=(XFGColor||SolidColor0)?XPal[XFGColor].ct_##type:bc;	\
								\
  M=0x80;							\
  CG=(UseFont&&FontBuf)? FontBuf:VRAM+CHRGEN(0,0x800);		\
  CMSK= ~COLTAB_MSK;  TMSK= ~CHRTAB_MSK(p,10);			\
  S=CG+((Y+VDP[23])&0x07);					\
  C=COLTAB((Y>>3)*10, 512);					\
  T=CHRTAB((Y>>3)*80, p, 4096);					\
  for ( X=80 ; X ; X--)						\
    {								\
      if (T&TMSK) T=CHRTAB((Y>>3)*80+(80-X), p, 4096);		\
      if (C&CMSK) C=COLTAB((Y>>3)*10+(80-X)/8, 512);		\
      K=*(S+((long)VRAM[T++]<<3));				\
      if (VRAM[C]&M)						\
	for (J=3 ; J-- ; K<<=2) 				\
	  {							\
	    *(buf++)=(K&0x80)?xfc:xbc;				\
	    *(buf++)=(K&0x40)?xfc:xbc;				\
	  }							\
      else							\
	for (J=3 ; J-- ; K<<=2) 				\
	  {							\
	    *(buf++)=(K&0x80)?fc:bc;				\
	    *(buf++)=(K&0x40)?fc:bc;				\
	  }							\
    }								\
}
#define PixLineTx80N(type)					\
static void PixLineTx80N_##type(int p, byte Y, void *Buf)	\
{								\
  byte X,J,K,M;							\
  type bc,fc,xbc,xfc;						\
  byte *CG, *S;							\
  register long T, TMSK, C, CMSK;				\
  type *buf;							\
								\
  buf=(type*)Buf;						\
  bc=XPal[BGColor].ct_##type;					\
  fc=(FGColor||SolidColor0)?XPal[FGColor].ct_##type:bc;      	\
  xbc=(XBGColor||SolidColor0)?XPal[XBGColor].ct_##type:bc;	\
  xfc=(XFGColor||SolidColor0)?XPal[XFGColor].ct_##type:bc;	\
								\
  M=0x80;							\
  CG=(UseFont&&FontBuf)? FontBuf:VRAM+CHRGEN(0,0x800);		\
  CMSK= ~COLTAB_MSK;  TMSK= ~CHRTAB_MSK(p,10);			\
  S=CG+((Y+VDP[23])&0x07);					\
  C=COLTAB((Y>>3)*10, 512);					\
  T=CHRTAB((Y>>3)*80, p, 4096);					\
  for ( X=80 ; X ; X--)						\
    {								\
      if (T&TMSK) T=CHRTAB((Y>>3)*80+(80-X), p, 4096);		\
      if (C&CMSK) C=COLTAB((Y>>3)*10+(80-X)/8, 512);		\
      K=*(S+((long)VRAM[T++]<<3));				\
      if (VRAM[C]&M)						\
	for (J=3 ; J-- ; K<<=2) 				\
	    *(buf++)=(K&0x80)?xfc:xbc;				\
      else							\
	for (J=3 ; J-- ; K<<=2) 				\
	    *(buf++)=(K&0x80)?fc:bc;				\
    }								\
}

BLOATIFY(PixLineTx80,(int p, byte Y, void *Buf))

void RefreshLineTx80(register byte Y)
{
  ColTyp BC;
  int p;

  if (!Mapped) return;
  INIPAL( (1<<BGColor)|(1<<FGColor)|(1<<XBGColor)|(1<<XFGColor) );
  BC=XPal[BGColor];
  if (ScreenON)
    { 
      BorderChange(Y,BC,BC,240,9);
      p= EVEN_ODD&&(INTLAC||eo_flag);
      if (postproc)
	{
	  PixLineTx80[PL_ind](p,Y,linbuf);
	  postproc(ScrPtr(Y),240);
	}
      else
	PixLineTx80[PL_ind](p,Y,ScrPtr(Y));
      if (high) 
	{
	  if (INTERLACING) 
	    {
	      if (postproc)
		{
		  PixLineTx80[PL_ind](0,Y,linbuf);
		  postproc(ScrPtr_Odd(Y),240);
		}
	      else
		PixLineTx80[PL_ind](0,Y,ScrPtr_Odd(Y));
	    }
	  else NoIntLac(Y);
	}
    }
  else BorderChange(Y,XPal[BGColor],XPal[BGColor],0,0);
}

void choosefuncs(int lsbfirst, int bitpix)
{
  int i;

  if (lsbfirst && (bitpix<8)) bitpix|=0x80;
  for (i=0 ; (PixSetters[i].bp_bitpix!=bitpix)&&(i<10) ; i++)
    ;
  postproc=PixSetters[i].bp_postproc;
  PL_ind=PixSetters[i].bp_index;
/* eo_flag=0; */
}

void setwidth(int wide)
{
  if (wide)
    PL_ind |= 1;
  else
    PL_ind &= ~1;
}

